"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tokenizer_1 = require("../../tokenizer");
const context_1 = require("../../tokenizer/context");
const types_1 = require("../../tokenizer/types");
const charCodes = require("../../util/charcodes");
const identifier_1 = require("../../util/identifier");
const whitespace_1 = require("../../util/whitespace");
const xhtml_1 = require("./xhtml");
const HEX_NUMBER = /^[\da-fA-F]+$/;
const DECIMAL_NUMBER = /^\d+$/;
context_1.types.j_oTag = new context_1.TokContext("<tag", false);
context_1.types.j_cTag = new context_1.TokContext("</tag", false);
context_1.types.j_expr = new context_1.TokContext("<tag>...</tag>", true, true);
types_1.types.jsxName = new types_1.TokenType("jsxName");
types_1.types.jsxText = new types_1.TokenType("jsxText", { beforeExpr: true });
types_1.types.jsxTagStart = new types_1.TokenType("jsxTagStart", { startsExpr: true });
types_1.types.jsxTagEnd = new types_1.TokenType("jsxTagEnd");
types_1.types.jsxTagStart.updateContext = function () {
    this.state.context.push(context_1.types.j_expr); // treat as beginning of JSX expression
    this.state.context.push(context_1.types.j_oTag); // start opening tag context
    this.state.exprAllowed = false;
};
types_1.types.jsxTagEnd.updateContext = function (prevType) {
    const out = this.state.context.pop();
    if ((out === context_1.types.j_oTag && prevType === types_1.types.slash) || out === context_1.types.j_cTag) {
        this.state.context.pop();
        this.state.exprAllowed = this.curContext() === context_1.types.j_expr;
    }
    else {
        this.state.exprAllowed = true;
    }
};
function isFragment(object) {
    return object
        ? object.type === "JSXOpeningFragment" || object.type === "JSXClosingFragment"
        : false;
}
// Transforms JSX element name to string.
function getQualifiedJSXName(object) {
    if (object.type === "JSXIdentifier") {
        return object.name;
    }
    if (object.type === "JSXNamespacedName") {
        return `${object.namespace.name}:${object.name.name}`;
    }
    if (object.type === "JSXMemberExpression") {
        return `${getQualifiedJSXName(object.object)}.${getQualifiedJSXName(object.property)}`;
    }
    // istanbul ignore next
    throw new Error(`Node had unexpected type: ${object.type}`);
}
exports.default = (superClass) => class extends superClass {
    // Reads inline JSX contents token.
    jsxReadToken() {
        let out = "";
        let chunkStart = this.state.pos;
        for (;;) {
            if (this.state.pos >= this.input.length) {
                this.raise(this.state.start, "Unterminated JSX contents");
            }
            const ch = this.input.charCodeAt(this.state.pos);
            switch (ch) {
                case charCodes.lessThan:
                case charCodes.leftCurlyBrace:
                    if (this.state.pos === this.state.start) {
                        if (ch === charCodes.lessThan && this.state.exprAllowed) {
                            ++this.state.pos;
                            this.finishToken(types_1.types.jsxTagStart);
                            return;
                        }
                        this.getTokenFromCode(ch);
                        return;
                    }
                    out += this.input.slice(chunkStart, this.state.pos);
                    this.finishToken(types_1.types.jsxText, out);
                    return;
                case charCodes.ampersand:
                    out += this.input.slice(chunkStart, this.state.pos);
                    out += this.jsxReadEntity();
                    chunkStart = this.state.pos;
                    break;
                default:
                    if (whitespace_1.isNewLine(ch)) {
                        out += this.input.slice(chunkStart, this.state.pos);
                        out += this.jsxReadNewLine(true);
                        chunkStart = this.state.pos;
                    }
                    else {
                        ++this.state.pos;
                    }
            }
        }
    }
    jsxReadNewLine(normalizeCRLF) {
        const ch = this.input.charCodeAt(this.state.pos);
        let out;
        ++this.state.pos;
        if (ch === charCodes.carriageReturn &&
            this.input.charCodeAt(this.state.pos) === charCodes.lineFeed) {
            ++this.state.pos;
            out = normalizeCRLF ? "\n" : "\r\n";
        }
        else {
            out = String.fromCharCode(ch);
        }
        ++this.state.curLine;
        this.state.lineStart = this.state.pos;
        return out;
    }
    jsxReadString(quote) {
        let out = "";
        let chunkStart = ++this.state.pos;
        for (;;) {
            if (this.state.pos >= this.input.length) {
                this.raise(this.state.start, "Unterminated string constant");
            }
            const ch = this.input.charCodeAt(this.state.pos);
            if (ch === quote)
                break;
            if (ch === charCodes.ampersand) {
                out += this.input.slice(chunkStart, this.state.pos);
                out += this.jsxReadEntity();
                chunkStart = this.state.pos;
            }
            else if (whitespace_1.isNewLine(ch)) {
                out += this.input.slice(chunkStart, this.state.pos);
                out += this.jsxReadNewLine(false);
                chunkStart = this.state.pos;
            }
            else {
                ++this.state.pos;
            }
        }
        out += this.input.slice(chunkStart, this.state.pos++);
        this.finishToken(types_1.types.string, out);
    }
    jsxReadEntity() {
        let str = "";
        let count = 0;
        let entity;
        let ch = this.input[this.state.pos];
        const startPos = ++this.state.pos;
        while (this.state.pos < this.input.length && count++ < 10) {
            ch = this.input[this.state.pos++];
            if (ch === ";") {
                if (str[0] === "#") {
                    if (str[1] === "x") {
                        str = str.substr(2);
                        if (HEX_NUMBER.test(str)) {
                            entity = String.fromCodePoint(parseInt(str, 16));
                        }
                    }
                    else {
                        str = str.substr(1);
                        if (DECIMAL_NUMBER.test(str)) {
                            entity = String.fromCodePoint(parseInt(str, 10));
                        }
                    }
                }
                else {
                    entity = xhtml_1.default[str];
                }
                break;
            }
            str += ch;
        }
        if (!entity) {
            this.state.pos = startPos;
            return "&";
        }
        return entity;
    }
    // Read a JSX identifier (valid tag or attribute name).
    //
    // Optimized version since JSX identifiers can"t contain
    // escape characters and so can be read as single slice.
    // Also assumes that first character was already checked
    // by isIdentifierStart in readToken.
    jsxReadWord() {
        let ch;
        const start = this.state.pos;
        do {
            ch = this.input.charCodeAt(++this.state.pos);
        } while (identifier_1.isIdentifierChar(ch) || ch === charCodes.dash);
        this.finishToken(types_1.types.jsxName, this.input.slice(start, this.state.pos));
    }
    // Parse next token as JSX identifier
    jsxParseIdentifier() {
        const node = this.startNode();
        if (this.match(types_1.types.jsxName)) {
            node.name = this.state.value;
        }
        else if (this.state.type.keyword) {
            node.name = this.state.type.keyword;
        }
        else {
            this.unexpected();
        }
        this.next();
        return this.finishNode(node, "JSXIdentifier");
    }
    // Parse namespaced identifier.
    jsxParseNamespacedName() {
        const startPos = this.state.start;
        const startLoc = this.state.startLoc;
        const name = this.jsxParseIdentifier();
        if (!this.eat(types_1.types.colon)) {
            // Plain identifier, so this is an access.
            this.state.tokens[this.state.tokens.length - 1].identifierRole = tokenizer_1.IdentifierRole.Access;
            return name;
        }
        const node = this.startNodeAt(startPos, startLoc);
        node.namespace = name;
        node.name = this.jsxParseIdentifier();
        return this.finishNode(node, "JSXNamespacedName");
    }
    // Parses element name in any form - namespaced, member
    // or single identifier.
    jsxParseElementName() {
        const startPos = this.state.start;
        const startLoc = this.state.startLoc;
        let node = this.jsxParseNamespacedName();
        while (this.eat(types_1.types.dot)) {
            const newNode = this.startNodeAt(startPos, startLoc);
            newNode.object = node;
            newNode.property = this.jsxParseIdentifier();
            node = this.finishNode(newNode, "JSXMemberExpression");
        }
        return node;
    }
    // Parses any type of JSX attribute value.
    jsxParseAttributeValue() {
        let node;
        switch (this.state.type) {
            case types_1.types.braceL:
                node = this.jsxParseExpressionContainer();
                if (node.expression.type === "JSXEmptyExpression") {
                    throw this.raise(node.start, "JSX attributes must only be assigned a non-empty expression");
                }
                else {
                    return node;
                }
            case types_1.types.jsxTagStart:
            case types_1.types.string:
                return this.parseExprAtom();
            default:
                throw this.raise(this.state.start, "JSX value should be either an expression or a quoted JSX text");
        }
    }
    // JSXEmptyExpression is unique type since it doesn't actually parse anything,
    // and so it should start at the end of last read token (left brace) and finish
    // at the beginning of the next one (right brace).
    jsxParseEmptyExpression() {
        const node = this.startNodeAt(this.state.lastTokEnd, this.state.lastTokEndLoc);
        return this.finishNodeAt(node, "JSXEmptyExpression", this.state.start, this.state.startLoc);
    }
    // Parse JSX spread child
    jsxParseSpreadChild() {
        const node = this.startNode();
        this.expect(types_1.types.braceL);
        this.expect(types_1.types.ellipsis);
        node.expression = this.parseExpression();
        this.expect(types_1.types.braceR);
        return this.finishNode(node, "JSXSpreadChild");
    }
    // Parses JSX expression enclosed into curly brackets.
    jsxParseExpressionContainer() {
        const node = this.startNode();
        this.next();
        if (this.match(types_1.types.braceR)) {
            node.expression = this.jsxParseEmptyExpression();
        }
        else {
            node.expression = this.parseExpression();
        }
        this.expect(types_1.types.braceR);
        return this.finishNode(node, "JSXExpressionContainer");
    }
    // Parses following JSX attribute name-value pair.
    jsxParseAttribute() {
        const node = this.startNode();
        if (this.eat(types_1.types.braceL)) {
            this.expect(types_1.types.ellipsis);
            node.argument = this.parseMaybeAssign();
            this.expect(types_1.types.braceR);
            return this.finishNode(node, "JSXSpreadAttribute");
        }
        node.name = this.jsxParseNamespacedName();
        node.value = this.eat(types_1.types.eq) ? this.jsxParseAttributeValue() : null;
        return this.finishNode(node, "JSXAttribute");
    }
    // Parses JSX opening tag starting after "<".
    jsxParseOpeningElementAt(startPos, startLoc) {
        const node = this.startNodeAt(startPos, startLoc);
        if (this.match(types_1.types.jsxTagEnd)) {
            this.expect(types_1.types.jsxTagEnd);
            return this.finishNode(node, "JSXOpeningFragment");
        }
        node.attributes = [];
        node.name = this.jsxParseElementName();
        while (!this.match(types_1.types.slash) && !this.match(types_1.types.jsxTagEnd)) {
            node.attributes.push(this.jsxParseAttribute());
        }
        node.selfClosing = this.eat(types_1.types.slash);
        this.expect(types_1.types.jsxTagEnd);
        return this.finishNode(node, "JSXOpeningElement");
    }
    // Parses JSX closing tag starting after "</".
    jsxParseClosingElementAt(startPos, startLoc) {
        const node = this.startNodeAt(startPos, startLoc);
        if (this.match(types_1.types.jsxTagEnd)) {
            this.expect(types_1.types.jsxTagEnd);
            return this.finishNode(node, "JSXClosingFragment");
        }
        node.name = this.jsxParseElementName();
        this.expect(types_1.types.jsxTagEnd);
        return this.finishNode(node, "JSXClosingElement");
    }
    // Parses entire JSX element, including it"s opening tag
    // (starting after "<"), attributes, contents and closing tag.
    jsxParseElementAt(startPos, startLoc) {
        const node = this.startNodeAt(startPos, startLoc);
        const children = [];
        const openingElement = this.jsxParseOpeningElementAt(startPos, startLoc);
        let closingElement = null;
        if (!openingElement.selfClosing) {
            contents: for (;;) {
                switch (this.state.type) {
                    case types_1.types.jsxTagStart:
                        startPos = this.state.start;
                        startLoc = this.state.startLoc;
                        this.next();
                        if (this.eat(types_1.types.slash)) {
                            closingElement = this.jsxParseClosingElementAt(startPos, startLoc);
                            break contents;
                        }
                        children.push(this.jsxParseElementAt(startPos, startLoc));
                        break;
                    case types_1.types.jsxText:
                        children.push(this.parseExprAtom());
                        break;
                    case types_1.types.braceL:
                        if (this.lookahead().type === types_1.types.ellipsis) {
                            children.push(this.jsxParseSpreadChild());
                        }
                        else {
                            children.push(this.jsxParseExpressionContainer());
                        }
                        break;
                    // istanbul ignore next - should never happen
                    default:
                        throw this.unexpected();
                }
            }
            if (isFragment(openingElement) && !isFragment(closingElement)) {
                this.raise(
                // $FlowIgnore
                closingElement.start, "Expected corresponding JSX closing tag for <>");
            }
            else if (!isFragment(openingElement) && isFragment(closingElement)) {
                this.raise(
                // $FlowIgnore
                closingElement.start, `Expected corresponding JSX closing tag for <${getQualifiedJSXName(openingElement.name)}>`);
            }
            else if (!isFragment(openingElement) && !isFragment(closingElement)) {
                if (
                // $FlowIgnore
                getQualifiedJSXName(closingElement.name) !== getQualifiedJSXName(openingElement.name)) {
                    this.raise(
                    // $FlowIgnore
                    closingElement.start, `Expected corresponding JSX closing tag for <${getQualifiedJSXName(openingElement.name)}>`);
                }
            }
        }
        if (isFragment(openingElement)) {
            node.openingFragment = openingElement;
            node.closingFragment = closingElement;
        }
        else {
            node.openingElement = openingElement;
            node.closingElement = closingElement;
        }
        node.children = children;
        if (this.match(types_1.types.relational) && this.state.value === "<") {
            this.raise(this.state.start, "Adjacent JSX elements must be wrapped in an enclosing tag");
        }
        return isFragment(openingElement)
            ? this.finishNode(node, "JSXFragment")
            : this.finishNode(node, "JSXElement");
    }
    // Parses entire JSX element from current position.
    jsxParseElement() {
        const startPos = this.state.start;
        const startLoc = this.state.startLoc;
        this.next();
        return this.jsxParseElementAt(startPos, startLoc);
    }
    // ==================================
    // Overrides
    // ==================================
    parseExprAtom(refShortHandDefaultPos = null) {
        if (this.match(types_1.types.jsxText)) {
            return this.parseLiteral(this.state.value, "JSXText");
        }
        else if (this.match(types_1.types.jsxTagStart)) {
            return this.jsxParseElement();
        }
        else {
            return super.parseExprAtom(refShortHandDefaultPos);
        }
    }
    readToken(code) {
        if (this.state.inPropertyName) {
            super.readToken(code);
            return;
        }
        const context = this.curContext();
        if (context === context_1.types.j_expr) {
            this.jsxReadToken();
            return;
        }
        if (context === context_1.types.j_oTag || context === context_1.types.j_cTag) {
            if (identifier_1.isIdentifierStart(code)) {
                this.jsxReadWord();
                return;
            }
            if (code === charCodes.greaterThan) {
                ++this.state.pos;
                this.finishToken(types_1.types.jsxTagEnd);
                return;
            }
            if ((code === charCodes.quotationMark || code === charCodes.apostrophe) &&
                context === context_1.types.j_oTag) {
                this.jsxReadString(code);
                return;
            }
        }
        if (code === charCodes.lessThan && this.state.exprAllowed) {
            ++this.state.pos;
            this.finishToken(types_1.types.jsxTagStart);
            return;
        }
        super.readToken(code);
    }
    updateContext(prevType) {
        if (this.match(types_1.types.braceL)) {
            const curContext = this.curContext();
            if (curContext === context_1.types.j_oTag) {
                this.state.context.push(context_1.types.braceExpression);
            }
            else if (curContext === context_1.types.j_expr) {
                this.state.context.push(context_1.types.templateQuasi);
            }
            else {
                super.updateContext(prevType);
            }
            this.state.exprAllowed = true;
        }
        else if (this.match(types_1.types.slash) && prevType === types_1.types.jsxTagStart) {
            this.state.context.length -= 2; // do not consider JSX expr -> JSX open tag -> ... anymore
            this.state.context.push(context_1.types.j_cTag); // reconsider as closing tag context
            this.state.exprAllowed = false;
        }
        else {
            super.updateContext(prevType);
        }
    }
};
